@@ -56,6 +56,9 @@ Capture = collections.namedtuple("Capture", "stdout stderr exit_status")
 def is_fileno(n, f):
   return (f is n) or (hasattr(f, 'fileno') and f.fileno() == n)
 
+def name_or_self(f):
+  return (hasattr(f, 'name') and f.name) or f
+
 
 class Cmd(object):
   def __init__(self, cmd, fd={}, e={}, cd=None):
@@ -123,7 +126,7 @@ class Cmd(object):
    
   def __repr__(self):
     return "Cmd(%r, fd=%r, e=%r, cd=%r)" % (self.cmd, dict(
-    			(k, v.name if isinstance(v, file) else v) for k, v in self.fd.iteritems()
+    			(k, name_or_self(v)) for k, v in self.fd.iteritems()
     		), self.e, self.cd)
   
   def __eq__(self, other):
@@ -188,12 +191,14 @@ class Cmd(object):
       if is_fileno(1, self.fd[1]):
         self.fd[1] = tempfile.TemporaryFile()
       else:
-        raise ValueError("cannot capture the child's stdout: it had been redirected to %s" % self.fd[1])
+        raise ValueError("cannot capture the child's stdout: it had been redirected to %r"
+        		% name_or_self(self.fd[1]))
     if 2 in fd:
       if is_fileno(2, self.fd[2]):
         self.fd[2] = tempfile.TemporaryFile()
       else:
-        raise ValueError("cannot capture the child's stderr: it had been redirected to %s" % self.fd[2])
+        raise ValueError("cannot capture the child's stderr: it had been redirected to %r"
+        		% name_or_self(self.fd[2]))
     p = subprocess.Popen(self.cmd, cwd=self.cd, env=self.env, stdin=self.fd[0], stdout=self.fd[1], stderr=self.fd[2])
     if p.stdin:
       p.stdin.close()
@@ -219,7 +224,7 @@ class Sh(Cmd):
   
   def __repr__(self):
     return "Sh(%r, fd=%r, e=%r, cd=%r)" % (self.cmd[2], dict(
-    			(k, v.name if isinstance(v, file) else v) for k, v in self.fd.iteritems()
+    			(k, name_or_self(v)) for k, v in self.fd.iteritems()
     		), self.e, self.cd)
 
 
@@ -254,13 +259,21 @@ class Pipe(object):
     """
     Fork-exec the pipeline and wait for its termination.
     
-    Return the last child's exit status.
+    Return an array of all children's exit status.
+    
+    >>> Pipe(Sh("echo foo"), Sh("cat; echo bar"), Cmd("cat", {1: os.devnull})).run()
+    [0, 0, 0]
     """
     prev = self.cmd[0].fd[0]
     for c in self.cmd:
       c.p = subprocess.Popen(c.cmd, stdin=prev, stdout=c.fd[1], stderr=c.fd[2], cwd=c.cd, env=c.env)
       prev = c.p.stdout
-    return self.cmd[-1].p.wait()
+    for c in self.cmd:
+      c.p.wait()
+    for c in self.cmd[:-1]:
+      if c.fd[1] == PIPE:
+        c.p.stdout.close()
+    return [c.p.returncode for c in self.cmd]
   
   def spawn(self):
     """
@@ -269,6 +282,8 @@ class Pipe(object):
     After spawned, each self.cmd[i] will have a 'p' attribute that is
     the spawned subprocess.Popen object.
     
+    Remember that all of [c.p.stdout for c in self.cmd] are open files.
+    
     >>> yesno = Pipe(Cmd('yes'), Cmd(['grep', 'no'])).spawn()
     >>> yesno.cmd[0].p.kill()
     >>> yesno.cmd[-1].p.wait()
@@ -311,10 +326,12 @@ class Pipe(object):
     if not fd <= set([1, 2]):
       raise NotImplementedError("can only capture a subset of fd [1, 2] for now")
     if 1 in fd and not is_fileno(1, self.fd[1]):
-      raise ValueError("cannot capture the last child's stdout: it had been redirected to %s" % self.fd[1])
+      raise ValueError("cannot capture the last child's stdout: it had been redirected to %r"
+      		% name_or_self(self.fd[1]))
     temp = None
     if 2 in fd:
       self.fd[2] = tempfile.TemporaryFile()
+    ## start piping
     prev = self.cmd[0].fd[0]
     for c in self.cmd[:-1]:
       if not is_fileno(0, c.fd[0]):
@@ -323,6 +340,7 @@ class Pipe(object):
         c.fd[2] = self.fd[2]
       c.p = subprocess.Popen(c.cmd, stdin=prev, stdout=c.fd[1], stderr=c.fd[2], cwd=c.cd, env=c.env)
       prev = c.p.stdout
+    ## prepare and fork the last child
     c = self.cmd[-1]
     if not is_fileno(0, c.fd[0]):
       prev = c.fd[0]
@@ -332,8 +350,13 @@ class Pipe(object):
     if 2 in fd and is_fileno(2, c.fd[2]):
       c.fd[2] = self.fd[2]
     c.p = subprocess.Popen(c.cmd, stdin=prev, stdout=c.fd[1], stderr=c.fd[2], cwd=c.cd, env=c.env)
+    ## wait for all children
     for c in self.cmd:
       c.p.wait()
+    ## close all unneeded files
+    for c in self.cmd[:-1]:
+      if c.fd[1] == PIPE:
+        c.p.stdout.close()
     if len(fd) == 1:
       if 1 in fd and not is_fileno(2, self.fd[2]): self.fd[2].close()
       if 2 in fd and not is_fileno(1, self.fd[1]): self.fd[1].close()
